use num_bigint::BigInt;
use num_traits::{One, Zero};

// Legendre symbol, returns 1, 0, or -1 mod p
fn ls(a: &BigInt, p: &BigInt) -> BigInt {
    let exp = (p - BigInt::one()) / BigInt::from(2);
    a.modpow(&exp, p)
}

// Tonelli-Shanks algorithm
pub fn mod_sqrt(n: &BigInt, p: &BigInt) -> Option<BigInt> {
    if !ls(n, p).is_one() {
        return None;
    }

    let mut q = p - 1;
    let mut s = BigInt::zero();
    while (&q & &BigInt::one()).is_zero() {
        s += 1;
        q >>= 1
    }

    if s.is_one() {
        let exp = (p + 1) / 4;
        let r1 = n.modpow(&exp, p);
        return Some(p - &r1);
    }

    let mut z = BigInt::from(2);
    while ls(&z, p) != p - 1 {
        z += 1
    }
    let mut c = z.modpow(&q, p);

    let mut r = n.modpow(&((&q + 1) / 2), p);
    let mut t = n.modpow(&q, p);
    let mut m = s;

    loop {
        if t.is_one() {
            return Some(p - &r);
        }

        let mut i = BigInt::zero();
        let mut z = t.clone();
        let mut b = c.clone();
        while !z.is_one() && i < &m - 1 {
            z = &z * &z % p;
            i += 1;
        }
        let mut e = &m - &i - 1;
        while e > BigInt::zero() {
            b = &b * &b % p;
            e -= 1;
        }
        r = &r * &b % p;
        c = &b * &b % p;
        t = &t * &c % p;
        m = i;
    }
}

#[cfg(test)]
mod tests {
    use openssl::bn::{BigNum, BigNumContext};

    #[test]
    fn test() {
        let n = "1573769205219068003359487454943505465350185201622247106224183162699789934914421854093895861489537435530695722027283028984396996570717144433549421890648379";
        let p = "5622613991231344109";

        {
            let n = num_bigint::BigInt::parse_bytes(n.as_bytes(), 10).unwrap();
            let p = num_bigint::BigInt::parse_bytes(p.as_bytes(), 10).unwrap();

            let r = super::mod_sqrt(&n, &p).unwrap();
            dbg!(r);
        }

        {
            let mut ctx = BigNumContext::new().unwrap();
            let n = BigNum::from_dec_str(n).unwrap();
            let p = BigNum::from_dec_str(p).unwrap();

            let mut r = BigNum::new().unwrap();
            r.mod_sqrt(&n, &p, &mut ctx).unwrap();
            dbg!(r.to_dec_str().unwrap());
        }
    }
}
